/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.shenyu.integratedtest.http;

import com.google.common.collect.Maps;
import com.google.gson.reflect.TypeToken;
import org.apache.shenyu.common.utils.SignUtils;
import org.apache.shenyu.integratedtest.http.dto.AdminResponse;
import org.apache.shenyu.integratedtest.http.dto.UserDTO;
import org.apache.shenyu.integratedtest.http.helper.HttpHelper;
import org.junit.Test;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.Map;
import java.util.concurrent.Future;
import static org.junit.Assert.assertEquals;

public class SignTest extends AbstractTest {

    @Test
    public void testSign() throws Exception {
        final String appKey = "108C27175A2C43C1BC29B1E483D57E3D";
        final String appSecret = "061521A73DD94A3FA873C25D050685BB";
        final String path = "/http/test/path/456";
        final String testUrlPath = "/http/test/path/456?name=Lee";
        final String version = "1.0.0";
        String now = String.valueOf(LocalDateTime.now().toInstant(ZoneOffset.of("+0")).toEpochMilli());
        Map<String, Object> normalHeaders = buildHeadrsMap(now, path, appKey, appSecret, version);
        UserDTO normalRespFuture = HttpHelper.INSTANCE.getFromGateway(testUrlPath,normalHeaders,
                UserDTO.class);
        assertEquals("hello world", normalRespFuture.getUserName());

        Map<String, Object> errorPathHeaders = buildHeadrsMap(now, "errorPath", appKey, appSecret, version);
        AdminResponse<Object> rejectedErrorPathRespFuture = HttpHelper.INSTANCE.getFromGateway(testUrlPath,errorPathHeaders,
                new TypeToken<AdminResponse<Object>>() {
                }.getType());
        assertEquals("signature value is error!",rejectedErrorPathRespFuture.getMessage());

        Map<String, Object> errorAppKeyHeaders = buildHeadrsMap(now, path, "ERRORKEY", appSecret, version);
        AdminResponse<Object> rejectedErrorAKRespFuture = HttpHelper.INSTANCE.getFromGateway(testUrlPath,errorAppKeyHeaders,
                new TypeToken<AdminResponse<Object>>() {
                }.getType());
        assertEquals("sign appKey does not exist.",rejectedErrorAKRespFuture.getMessage());

        Map<String, Object> errorAppSecretHeaders = buildHeadrsMap(now, path, appKey, "ERRORSECRET", version);
        AdminResponse<Object> rejectedErrorSKRespFuture = HttpHelper.INSTANCE.getFromGateway(testUrlPath,errorAppSecretHeaders,
                new TypeToken<AdminResponse<Object>>() {
                }.getType());
        assertEquals("signature value is error!",rejectedErrorSKRespFuture.getMessage());

        Map<String, Object> errorVersionHeaders = buildHeadrsMap(now, path, appKey, appSecret, "1.0.2");
        AdminResponse<Object> rejectedErrorVersionRespFuture = HttpHelper.INSTANCE.getFromGateway(testUrlPath,errorVersionHeaders,
                new TypeToken<AdminResponse<Object>>() {
                }.getType());
        assertEquals("signature value is error!",rejectedErrorVersionRespFuture.getMessage());

        String errorTime = String.valueOf(LocalDateTime.now().toInstant(ZoneOffset.of("+0")).toEpochMilli()-360000);
        Map<String, Object> errorTimestampHeaders = buildHeadrsMap(errorTime, path, appKey, appSecret, version);
        AdminResponse<Object> rejectedErrorTimestampRespFuture = HttpHelper.INSTANCE.getFromGateway(testUrlPath,errorTimestampHeaders,
                new TypeToken<AdminResponse<Object>>() {
                }.getType());
        assertEquals("The signature timestamp has exceeded 5 minutes!",rejectedErrorTimestampRespFuture.getMessage());
    }

    private Map<String, Object> buildHeadrsMap(final String timestamp,final String path,final String appKey,final String appSecret,final String version) {
        Map<String, String> params = Maps.newHashMapWithExpectedSize(3);
        params.put("timestamp",timestamp);
        params.put("path", path);
        params.put("version", version);
        String sign = SignUtils.generateSign(appSecret, params);

        Map<String, Object> headers = Maps.newHashMapWithExpectedSize(4);
        headers.put("timestamp",timestamp);
        headers.put("appKey",appKey);
        headers.put("sign",sign);
        headers.put("version",version);
        return headers;
    }
}
